use std::cell::RefCell;
use std::collections::BTreeMap;
use std::cmp::{ PartialOrd, Ord };

#[derive(Default)]
pub struct Memo<I, O> {
    memoized: RefCell<BTreeMap<I, O>>,
}

impl<I, O> Memo<I, O>
where
    I: Clone + PartialOrd + Ord,
    O: Clone,
{
    pub fn memoize<F>(&self, input: I, work: F) -> O
    where
        F: Fn(I) -> O,
    {
        let memoized_ref = self.memoized.borrow();
        match memoized_ref.get(&input) {
            Some(memoized) => memoized.clone(),
            None => {
                drop(memoized_ref);
                let worked = work(input.clone());
                let mut memoized_ref = self.memoized.borrow_mut();
                memoized_ref.insert(input, worked.clone());
                worked
            },
        }
    }
}

macro_rules! memoize {
    (
        $(#[$meta:meta])*
        $vis:vis fn $fn:ident
        $(<
            $($lt:lifetime,)*
            $($gt_ident:ident$(: $gt_ty:ty)?,)*
            $(const $c_ident:ident: $c_ty:ty,)*
        >)?
        ($($a_ident:ident: $a_ty:ty),* $(,)?)
        -> $r_ty:ty
        $body:block
    ) => {
        $(#[$meta])*
        $vis fn $fn
        $(<$($lt,)* $($gt_ident$(: $gt_ty)?,)* $($c_ident: $c_ty,)*>)?
        ($($a_ident: $a_ty),*)
        -> $r_ty
        {
            thread_local! {
                static MEMO: crate::memo::Memo<($($a_ty,)*), $r_ty> = crate::memo::Memo::default();
            }
            MEMO.with(|memo| memo.memoize(($($a_ident,)*), |($($a_ident,)*)| $body))
        }
    };
}
